/*
 * See the dyninst/COPYRIGHT file for copyright information.
 * 
 * We provide the Paradyn Tools (below described as "Paradyn")
 * on an AS IS basis, and do not warrant its validity or performance.
 * We reserve the right to update, modify, or discontinue this
 * software at any time.  We shall have no obligation to supply such
 * updates or modifications or any other form of support to you.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#ifndef _STRING_DECL_H
#define _STRING_DECL_H

/*
 * $Log: stringDecl.h,v $
 * Revision 1.6  2007/05/30 19:20:07  legendre
 * Update copyright notice to LGPL
 *
 * Revision 1.5  2005/12/19 19:41:40  pack
 * MRNet replaces xdr for frontend to daemon communication.  New igen  - - -
 * MRNet changes
 *
 * Revision 1.1.1.1  2004/11/05 22:17:36  darnold
 * start
 *
 * Revision 1.1.1.1  2004/10/18 20:35:55  darnold
 * Manual Branch for MRNet Integration based on Release_4.1
 *
 * Revision 1.4  2004/03/23 01:11:53  eli
 * Updated copyright string
 *
 * Revision 1.3  1996/08/16 21:30:46  tamches
 * updated copyright for release 1.1
 *
 * Revision 1.2  1994/09/02 19:42:46  markc
 * Added missing semicolon to end of typedef
 *
 * Revision 1.1  1994/09/02  10:43:50  markc
 * Moved typedef for stringHandle outside of stringPool.h
 *
 */

typedef void *stringHandle;

#endif /* _STRING_DECL_H */
