/*
 * See the dyninst/COPYRIGHT file for copyright information.
 * 
 * We provide the Paradyn Tools (below described as "Paradyn")
 * on an AS IS basis, and do not warrant its validity or performance.
 * We reserve the right to update, modify, or discontinue this
 * software at any time.  We shall have no obligation to supply such
 * updates or modifications or any other form of support to you.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#if !defined(_Resume_Log_h_)
#define _Resume_Log_h_

#include <vector>
#include "test_info_new.h"

TESTLIB_DLL_EXPORT void enableResumeLog();
TESTLIB_DLL_EXPORT bool isLogging();
TESTLIB_DLL_EXPORT void setLoggingFilename(char *f);

TESTLIB_DLL_EXPORT void log_teststart(int groupnum, int testnum, test_runstate_t runstate);
TESTLIB_DLL_EXPORT void log_testresult(test_results_t result);
TESTLIB_DLL_EXPORT void log_testreported(int groupnum, int testnum);
TESTLIB_DLL_EXPORT void log_resumepoint(int groupnum, int testnum);
TESTLIB_DLL_EXPORT void log_clear();
TESTLIB_DLL_EXPORT void parse_resumelog(std::vector<RunGroup *> &groups);

TESTLIB_DLL_EXPORT void parse_mutateelog(RunGroup *group, char *logname);
TESTLIB_DLL_EXPORT void clear_mutateelog(char *logname);

TESTLIB_DLL_EXPORT char *get_resumelog_name();
TESTLIB_DLL_EXPORT void set_resumelog_name(char *s);

#endif
