/* This file automatically generated from test specifications.  See
 * specification/spec.pl and specification/makemake.py
 */

#include "test_info_new.h"



static unsigned int test_count = 0;
static unsigned int group_count = 0;
static std::vector<RunGroup *> *tests = NULL;

static void fini_group(RunGroup *rg) {
  rg->index = group_count++;
  tests->push_back(rg);
  test_count = 0;
}

static void add_test(RunGroup *rg, const char *ts) {
  rg->tests.push_back(new TestInfo(test_count++, ".so", ts));
}

// Now we insert the test lists into the run groups
void initialize_mutatees(std::vector<RunGroup *> &t) {
        tests = &t;
	RunGroup *rg;
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_static_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_cxx_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_6, mutator: test5_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_static_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_35.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_35, mutator: test1_35, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_35, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_1.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_2.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_3.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_4.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_5.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_6.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_7.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_mem_8.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_1.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_1, mutator: test_thread_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_2.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_2, mutator: test_thread_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_3.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_3, mutator: test_thread_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_5.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_5, mutator: test_thread_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_pic_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_gcc_32_none_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_pic_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_6.mutatee_solo_dynamic_g++_32_none_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_pic_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_none_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_pic_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_none_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_pic_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_none_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_pic_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_none_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_pic_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_none_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_pic_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_gcc_32_none_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_pic_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_none_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_pic_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_none_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_pic_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_none_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_pic_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_none_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_pic_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_none_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_pic_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_7.mutatee_solo_dynamic_g++_32_none_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_pic_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_gcc_32_none_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_none", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_none", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_low", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_low", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_high", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_high", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_max", DELAYEDATTACH, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_pic_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: pic, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_thread_8.mutatee_solo_dynamic_g++_32_none_max", DELAYEDATTACH, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_launch.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread_cont.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_breakpoint.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_library.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_addlibrary.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_singlestep.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_thread.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_groups.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_stat.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_irpc.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_temp_detach.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_gcc_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, MultiThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, MultiProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, CREATE, SingleThreaded, SingleProcess, local, local, no_launch, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_pic_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, PIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: pic, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, MultiThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, MultiProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("pc_terminate_stopped.mutatee_solo_dynamic_g++_32_none_none", SELFATTACH, USEATTACH, SingleThreaded, SingleProcess, local, local, pre, DynamicLink, true, nonPIC, "proccontrol", "g++", "none", "32", "NONE");
  add_test(rg, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, DISK, TNone, PNone, local, local, post, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test_pt_ls, mutator: test_pt_ls, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: fucompp, mutator: fucompp, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: stack_mov_32_bit, mutator: stack_mov_32_bit, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: mov_size_details, mutator: mov_size_details, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test_pt_ls, mutator: test_pt_ls, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test2_1, mutator: test2_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test2_2, mutator: test2_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test_pt_ls, mutator: test_pt_ls, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test2_3, mutator: test2_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("", SELFSTART, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "", "none", "32", "NONE");
  add_test(rg, "{test: test2_4, mutator: test2_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: test_instruction_read_write, mutator: test_instruction_read_write, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: test_instruction_farcall, mutator: test_instruction_farcall, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: test_instruction_bind_eval, mutator: test_instruction_bind_eval, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: test_instruction_profile, mutator: test_instruction_profile, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: power_decode, mutator: power_decode, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "instruction", "", "none", "32", "NONE");
  add_test(rg, "{test: power_cft, mutator: power_cft, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_static_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("dyninst_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_39, mutator: test1_39, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  add_test(rg, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "symtab", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("symtab_group_test.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "symtab", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_add_symbols, mutator: test_add_symbols, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_relocations, mutator: test_relocations, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  add_test(rg, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: disk}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_12.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_static_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_14.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_19.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test_snip_remove.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("init_fini_callback.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: init_fini_callback, mutator: init_fini_callback, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: init_fini_callback, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_none", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_low", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_high", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_static_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, StaticLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: staticMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_max", STOPPED, DISK, TNone, PNone, local, local, post, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: post, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: binary}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_29.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_40.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, false, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_40, mutator: test1_40, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_40, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test1_41.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test1_41, mutator: test1_41, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_41, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_6.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_6, mutator: test2_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_8.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_none", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_low", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_high", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test2_14.mutatee_solo_dynamic_g++_32_none_max", STOPPED, USEATTACH, TNone, PNone, local, local, pre, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: useAttach}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_1.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_2.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_3.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_4.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_5.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_5, mutator: test3_5, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_6.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_6, mutator: test3_6, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test3_7.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_1.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_2.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_2, mutator: test4_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_3.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_3, mutator: test4_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_gcc_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_pic_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_none_none", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_pic_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_none_low", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_pic_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_none_high", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_pic_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: pic, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test4_4.mutatee_solo_dynamic_g++_32_none_max", SELFSTART, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test4_4, mutator: test4_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_5.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_5, mutator: test_fork_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_6.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_6, mutator: test_fork_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_7.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_7, mutator: test_fork_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_8.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_8, mutator: test_fork_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_9.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_9, mutator: test_fork_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_9, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_10.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_10, mutator: test_fork_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_10, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_11.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_11, mutator: test_fork_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_11, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_12.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_12, mutator: test_fork_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_fork_13.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_fork_13, mutator: test_fork_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_fork_13, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_1.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_2.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_2, mutator: test_stack_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_3.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_stack_4.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_stack_4, mutator: test_stack_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_1.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_1, mutator: test_callback_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_gcc_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "gcc", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: gcc, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_pic_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_none_none", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "none", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_pic_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_none_low", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "low", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_pic_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_none_high", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "high", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_pic_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, PIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: pic, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
  rg = new RunGroup("test_callback_2.mutatee_solo_dynamic_g++_32_none_max", STOPPED, CREATE, TNone, PNone, local, local, no_launch, DynamicLink, true, nonPIC, "dyninst", "g++", "max", "32", "NONE");
  add_test(rg, "{test: test_callback_2, mutator: test_callback_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_callback_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: g++, run_mode: createProcess}");
  fini_group(rg);
}
