/* This file automatically generated from test specifications.  See
 * specification/spec.pl and specification/makemake.py
 */

#include "test_info_new.h"
void initialize_mutatees_dyninst(std::vector<RunGroup *> &tests);
void initialize_mutatees_symtab(std::vector<RunGroup *> &tests);
void initialize_mutatees_instruction(std::vector<RunGroup *> &tests);
void initialize_mutatees_proccontrol(std::vector<RunGroup *> &tests);
void initialize_mutatees(std::vector<RunGroup *> &tests) {  initialize_mutatees_dyninst(tests);  initialize_mutatees_symtab(tests);  initialize_mutatees_instruction(tests);  initialize_mutatees_proccontrol(tests);}
// Now we insert the test lists into the run groups
void initialize_mutatees_dyninst(std::vector<RunGroup *> &tests) {
	unsigned int group_count = 0;
	// Keep track of which element each test is, for later use with the resumelog
	unsigned int test_count;
	RunGroup *rg;
struct {

    char * mutatee_name;
    start_state_t state_init;
    create_mode_t attach_init;
    bool ex;
    bool presencevar;
    char* module;
    char* compiler;
    char* optimization;
    char* abi;
    test_pictype_t pic;
  } rungroup_params[] = { {"dyninst_cxx_group_test_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"dyninst_cxx_group_test_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_1_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_2_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_3_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_4_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_5_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_6_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_7_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_mem_8_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_none.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_none.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_low.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_low.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_high.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_high.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_max.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC_32_max.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_none.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_none.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_low.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_low.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_high.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_high.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_max.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_thread_6_mutatee_solo_VC++_32_max.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_none.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_none.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_low.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_low.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_high.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_high.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_max.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC_32_max.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_none.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_none.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_low.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_low.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_high.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_high.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_max.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_thread_7_mutatee_solo_VC++_32_max.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_none.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_none.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_low.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_low.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_high.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_high.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_max.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC_32_max.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_none.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_none.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_low.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_low.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_high.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_high.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_max.exe", DELAYEDATTACH, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_thread_8_mutatee_solo_VC++_32_max.exe", DELAYEDATTACH, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"", SELFSTART, CREATE, true, true, "dyninst", "", "none", "32", nonPIC},
 {"", SELFSTART, CREATE, true, true, "dyninst", "", "none", "32", nonPIC},
 {"", SELFSTART, USEATTACH, true, true, "dyninst", "", "none", "32", nonPIC},
 {"", SELFSTART, USEATTACH, true, true, "dyninst", "", "none", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"dyninst_group_test_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_12_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_12_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_14_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_14_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_19_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_19_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_snip_remove_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_29_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test1_29_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test2_8_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test2_8_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test2_14_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test2_14_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test3_1_mutatee_solo_VC_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test3_1_mutatee_solo_VC_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test3_1_mutatee_solo_VC_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test3_1_mutatee_solo_VC_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test3_1_mutatee_solo_VC++_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test3_1_mutatee_solo_VC++_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test3_1_mutatee_solo_VC++_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test3_1_mutatee_solo_VC++_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test3_2_mutatee_solo_VC_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test3_2_mutatee_solo_VC_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test3_2_mutatee_solo_VC_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test3_2_mutatee_solo_VC_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test3_2_mutatee_solo_VC++_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test3_2_mutatee_solo_VC++_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test3_2_mutatee_solo_VC++_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test3_2_mutatee_solo_VC++_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test3_3_mutatee_solo_VC_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test3_3_mutatee_solo_VC_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test3_3_mutatee_solo_VC_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test3_3_mutatee_solo_VC_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test3_3_mutatee_solo_VC++_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test3_3_mutatee_solo_VC++_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test3_3_mutatee_solo_VC++_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test3_3_mutatee_solo_VC++_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test3_4_mutatee_solo_VC_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test3_4_mutatee_solo_VC_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test3_4_mutatee_solo_VC_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test3_4_mutatee_solo_VC_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test3_4_mutatee_solo_VC++_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test3_4_mutatee_solo_VC++_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test3_4_mutatee_solo_VC++_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test3_4_mutatee_solo_VC++_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test3_7_mutatee_solo_VC_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test3_7_mutatee_solo_VC_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test3_7_mutatee_solo_VC_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test3_7_mutatee_solo_VC_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test3_7_mutatee_solo_VC++_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test3_7_mutatee_solo_VC++_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test3_7_mutatee_solo_VC++_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test3_7_mutatee_solo_VC++_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test4_1_mutatee_solo_VC_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test4_1_mutatee_solo_VC_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test4_1_mutatee_solo_VC_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test4_1_mutatee_solo_VC_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test4_1_mutatee_solo_VC++_32_none.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test4_1_mutatee_solo_VC++_32_low.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test4_1_mutatee_solo_VC++_32_high.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test4_1_mutatee_solo_VC++_32_max.exe", SELFSTART, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_stack_1_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "none", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "low", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "high", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC", "max", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC++_32_none.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "none", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC++_32_low.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "low", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC++_32_high.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "high", "32", nonPIC},
 {"test_stack_3_mutatee_solo_VC++_32_max.exe", STOPPED, CREATE, true, true, "dyninst", "VC++", "max", "32", nonPIC} };

  struct {
    bool endrungroup;
    const char * iname;
    const char * mrname;
    const char * isoname;
	bool serialize_enable;
    const char * ilabel;
  } test_params[] = { {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test5_1", "test5_1", "test5_1.dll", false, "{test: test5_1, mutator: test5_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_2", "test5_2", "test5_2.dll", false, "{test: test5_2, mutator: test5_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_3", "test5_3", "test5_3.dll", false, "{test: test5_3, mutator: test5_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_4", "test5_4", "test5_4.dll", false, "{test: test5_4, mutator: test5_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_5", "test5_5", "test5_5.dll", false, "{test: test5_5, mutator: test5_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_7", "test5_7", "test5_7.dll", false, "{test: test5_7, mutator: test5_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test5_8", "test5_8", "test5_8.dll", false, "{test: test5_8, mutator: test5_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test5_9", "test5_9", "test5_9.dll", false, "{test: test5_9, mutator: test5_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_cxx_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_1", "test_mem_1", "test_mem_1.dll", false, "{test: test_mem_1, mutator: test_mem_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_2", "test_mem_2", "test_mem_2.dll", false, "{test: test_mem_2, mutator: test_mem_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_3", "test_mem_3", "test_mem_3.dll", false, "{test: test_mem_3, mutator: test_mem_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_4", "test_mem_4", "test_mem_4.dll", false, "{test: test_mem_4, mutator: test_mem_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_5", "test_mem_5", "test_mem_5.dll", false, "{test: test_mem_5, mutator: test_mem_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_5, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_6", "test_mem_6", "test_mem_6.dll", false, "{test: test_mem_6, mutator: test_mem_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_7", "test_mem_7", "test_mem_7.dll", false, "{test: test_mem_7, mutator: test_mem_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_mem_8", "test_mem_8", "test_mem_8.dll", false, "{test: test_mem_8, mutator: test_mem_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_mem_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_6", "test_thread_6", "test_thread_6.dll", false, "{test: test_thread_6, mutator: test_thread_6, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_6, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_7", "test_thread_7", "test_thread_7.dll", false, "{test: test_thread_7, mutator: test_thread_7, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_thread_8", "test_thread_8", "test_thread_8.dll", false, "{test: test_thread_8, mutator: test_thread_8, grouped: false, pic: none, start_state: delayedattach, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_thread_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_1", "test2_1", "test2_1.dll", false, "{test: test2_1, mutator: test2_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}"},
 {true, "test2_2", "test2_2", "test2_2.dll", false, "{test: test2_2, mutator: test2_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}"},
 {true, "test2_3", "test2_3", "test2_3.dll", false, "{test: test2_3, mutator: test2_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: useAttach}"},
 {true, "test2_4", "test2_4", "test2_4.dll", false, "{test: test2_4, mutator: test2_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {false, "snip_change_shlib_var", "snip_change_shlib_var", "snip_change_shlib_var.dll", false, "{test: snip_change_shlib_var, mutator: snip_change_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "snip_ref_shlib_var", "snip_ref_shlib_var", "snip_ref_shlib_var.dll", false, "{test: snip_ref_shlib_var, mutator: snip_ref_shlib_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_1", "test1_1", "test1_1.dll", false, "{test: test1_1, mutator: test1_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_10", "test1_10", "test1_10.dll", false, "{test: test1_10, mutator: test1_10, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_11", "test1_11", "test1_11.dll", false, "{test: test1_11, mutator: test1_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_13", "test1_13", "test1_13.dll", false, "{test: test1_13, mutator: test1_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_16", "test1_16", "test1_16.dll", false, "{test: test1_16, mutator: test1_16, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_17", "test1_17", "test1_17.dll", false, "{test: test1_17, mutator: test1_17, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_18", "test1_18", "test1_18.dll", false, "{test: test1_18, mutator: test1_18, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_2", "test1_2", "test1_2.dll", false, "{test: test1_2, mutator: test1_2, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_20", "test1_20", "test1_20.dll", false, "{test: test1_20, mutator: test1_20, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_21", "test1_21", "test1_21.dll", false, "{test: test1_21, mutator: test1_21, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_22", "test1_22", "test1_22.dll", false, "{test: test1_22, mutator: test1_22, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_23", "test1_23", "test1_23.dll", false, "{test: test1_23, mutator: test1_23, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_24", "test1_24", "test1_24.dll", false, "{test: test1_24, mutator: test1_24, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_25", "test1_25", "test1_25.dll", false, "{test: test1_25, mutator: test1_25, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_26", "test1_26", "test1_26.dll", false, "{test: test1_26, mutator: test1_26, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_27", "test1_27", "test1_27.dll", false, "{test: test1_27, mutator: test1_27, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_28", "test1_28", "test1_28.dll", false, "{test: test1_28, mutator: test1_28, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_3", "test1_3", "test1_3.dll", false, "{test: test1_3, mutator: test1_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_30", "test1_30", "test1_30.dll", false, "{test: test1_30, mutator: test1_30, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_31", "test1_31", "test1_31.dll", false, "{test: test1_31, mutator: test1_31, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_32", "test1_32", "test1_32.dll", false, "{test: test1_32, mutator: test1_32, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_33", "test1_33", "test1_33.dll", false, "{test: test1_33, mutator: test1_33, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_34", "test1_34", "test1_34.dll", false, "{test: test1_34, mutator: test1_34, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_36", "test1_36", "test1_36.dll", false, "{test: test1_36, mutator: test1_36, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_37", "test1_37", "test1_37.dll", false, "{test: test1_37, mutator: test1_37, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_38", "test1_38", "test1_38.dll", false, "{test: test1_38, mutator: test1_38, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_4", "test1_4", "test1_4.dll", false, "{test: test1_4, mutator: test1_4, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_5", "test1_5", "test1_5.dll", false, "{test: test1_5, mutator: test1_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_6", "test1_6", "test1_6.dll", false, "{test: test1_6, mutator: test1_6, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_7", "test1_7", "test1_7.dll", false, "{test: test1_7, mutator: test1_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_8", "test1_8", "test1_8.dll", false, "{test: test1_8, mutator: test1_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test1_9", "test1_9", "test1_9.dll", false, "{test: test1_9, mutator: test1_9, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_11", "test2_11", "test2_11.dll", false, "{test: test2_11, mutator: test2_11, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_12", "test2_12", "test2_12.dll", false, "{test: test2_12, mutator: test2_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_13", "test2_13", "test2_13.dll", false, "{test: test2_13, mutator: test2_13, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_5", "test2_5", "test2_5.dll", false, "{test: test2_5, mutator: test2_5, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {false, "test2_7", "test2_7", "test2_7.dll", false, "{test: test2_7, mutator: test2_7, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_write_param", "test_write_param", "test_write_param.dll", false, "{test: test_write_param, mutator: test_write_param, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: dyninst_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_12", "test1_12", "test1_12.dll", false, "{test: test1_12, mutator: test1_12, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_12, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_14", "test1_14", "test1_14.dll", false, "{test: test1_14, mutator: test1_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_19", "test1_19", "test1_19.dll", false, "{test: test1_19, mutator: test1_19, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_19, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_snip_remove", "test_snip_remove", "test_snip_remove.dll", false, "{test: test_snip_remove, mutator: test_snip_remove, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test_snip_remove, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test1_29", "test1_29", "test1_29.dll", false, "{test: test1_29, mutator: test1_29, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test1_29, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_8", "test2_8", "test2_8.dll", false, "{test: test2_8, mutator: test2_8, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_8, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test2_14", "test2_14", "test2_14.dll", false, "{test: test2_14, mutator: test2_14, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: pre, platmode: NONE, mutatee: test2_14, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_1", "test3_1", "test3_1.dll", false, "{test: test3_1, mutator: test3_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_2", "test3_2", "test3_2.dll", false, "{test: test3_2, mutator: test3_2, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_2, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_3", "test3_3", "test3_3.dll", false, "{test: test3_3, mutator: test3_3, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_4", "test3_4", "test3_4.dll", false, "{test: test3_4, mutator: test3_4, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_4, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test3_7", "test3_7", "test3_7.dll", false, "{test: test3_7, mutator: test3_7, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test3_7, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test4_1", "test4_1", "test4_1.dll", false, "{test: test4_1, mutator: test4_1, grouped: false, pic: none, start_state: selfstart, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test4_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_1", "test_stack_1", "test_stack_1.dll", false, "{test: test_stack_1, mutator: test_stack_1, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_1, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "test_stack_3", "test_stack_3", "test_stack_3.dll", false, "{test: test_stack_3, mutator: test_stack_3, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: test_stack_3, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: createProcess}"} };

  int tp_index = 0;
  for (int i = 0; i < 380; i++) {
    test_count = 0;
    rg = new RunGroup(rungroup_params[i].mutatee_name, rungroup_params[i].state_init, rungroup_params[i].attach_init, 
			rungroup_params[i].ex, rungroup_params[i].module, rungroup_params[i].pic, rungroup_params[i].compiler,
			rungroup_params[i].optimization, rungroup_params[i].abi, "NONE");
    
    for (; tp_index < 1060; tp_index++) {
      rg->tests.push_back(new TestInfo(test_count++, 
                                       test_params[tp_index].iname, 
                                       test_params[tp_index].mrname, 
                                       test_params[tp_index].isoname, 
                                       test_params[tp_index].serialize_enable, 
                                       test_params[tp_index].ilabel));
      if (test_params[tp_index].endrungroup == true) break;
    } 
    // Increment the test_params index (tp_index) past the endgroup. 
    tp_index++;
    rg->index = group_count++;
    tests.push_back(rg);
  }
}

// Now we insert the test lists into the run groups
void initialize_mutatees_symtab(std::vector<RunGroup *> &tests) {
	unsigned int group_count = 0;
	// Keep track of which element each test is, for later use with the resumelog
	unsigned int test_count;
	RunGroup *rg;
struct {

    char * mutatee_name;
    start_state_t state_init;
    create_mode_t attach_init;
    bool ex;
    bool presencevar;
    char* module;
    char* compiler;
    char* optimization;
    char* abi;
    test_pictype_t pic;
  } rungroup_params[] = { {"symtab_group_test_mutatee_solo_VC_32_none.exe", STOPPED, USEATTACH, false, true, "symtab", "VC", "none", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC_32_low.exe", STOPPED, USEATTACH, false, true, "symtab", "VC", "low", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC_32_high.exe", STOPPED, USEATTACH, false, true, "symtab", "VC", "high", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC_32_max.exe", STOPPED, USEATTACH, false, true, "symtab", "VC", "max", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC++_32_none.exe", STOPPED, USEATTACH, false, true, "symtab", "VC++", "none", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC++_32_low.exe", STOPPED, USEATTACH, false, true, "symtab", "VC++", "low", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC++_32_high.exe", STOPPED, USEATTACH, false, true, "symtab", "VC++", "high", "32", nonPIC},
 {"symtab_group_test_mutatee_solo_VC++_32_max.exe", STOPPED, USEATTACH, false, true, "symtab", "VC++", "max", "32", nonPIC} };

  struct {
    bool endrungroup;
    const char * iname;
    const char * mrname;
    const char * isoname;
	bool serialize_enable;
    const char * ilabel;
  } test_params[] = { {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: low, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: high, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_anno_basic_types", "test_anno_basic_types", "test_anno_basic_types.dll", false, "{test: test_anno_basic_types, mutator: test_anno_basic_types, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_line_info", "test_line_info", "test_line_info.dll", false, "{test: test_line_info, mutator: test_line_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_func", "test_lookup_func", "test_lookup_func.dll", false, "{test: test_lookup_func, mutator: test_lookup_func, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_lookup_var", "test_lookup_var", "test_lookup_var.dll", false, "{test: test_lookup_var, mutator: test_lookup_var, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_module", "test_module", "test_module.dll", false, "{test: test_module, mutator: test_module, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_ser_anno", "test_ser_anno", "test_ser_anno.dll", false, "{test: test_ser_anno, mutator: test_ser_anno, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {false, "test_symtab_ser_funcs", "test_symtab_ser_funcs", "test_symtab_ser_funcs.dll", false, "{test: test_symtab_ser_funcs, mutator: test_symtab_ser_funcs, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"},
 {true, "test_type_info", "test_type_info", "test_type_info.dll", false, "{test: test_type_info, mutator: test_type_info, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: symtab_group_test, mutatorstart: local, optimization: max, mutateestart: local, compiler: VC++, run_mode: disk}"} };

  int tp_index = 0;
  for (int i = 0; i < 8; i++) {
    test_count = 0;
    rg = new RunGroup(rungroup_params[i].mutatee_name, rungroup_params[i].state_init, rungroup_params[i].attach_init, 
			rungroup_params[i].ex, rungroup_params[i].module, rungroup_params[i].pic, rungroup_params[i].compiler,
			rungroup_params[i].optimization, rungroup_params[i].abi, "NONE");
    
    for (; tp_index < 64; tp_index++) {
      rg->tests.push_back(new TestInfo(test_count++, 
                                       test_params[tp_index].iname, 
                                       test_params[tp_index].mrname, 
                                       test_params[tp_index].isoname, 
                                       test_params[tp_index].serialize_enable, 
                                       test_params[tp_index].ilabel));
      if (test_params[tp_index].endrungroup == true) break;
    } 
    // Increment the test_params index (tp_index) past the endgroup. 
    tp_index++;
    rg->index = group_count++;
    tests.push_back(rg);
  }
}

// Now we insert the test lists into the run groups
void initialize_mutatees_instruction(std::vector<RunGroup *> &tests) {
	unsigned int group_count = 0;
	// Keep track of which element each test is, for later use with the resumelog
	unsigned int test_count;
	RunGroup *rg;
struct {

    char * mutatee_name;
    start_state_t state_init;
    create_mode_t attach_init;
    bool ex;
    bool presencevar;
    char* module;
    char* compiler;
    char* optimization;
    char* abi;
    test_pictype_t pic;
  } rungroup_params[] = { {"", STOPPED, CREATE, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, CREATE, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, CREATE, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, USEATTACH, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, USEATTACH, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, USEATTACH, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, USEATTACH, true, true, "instruction", "", "none", "32", nonPIC},
 {"", STOPPED, USEATTACH, true, true, "instruction", "", "none", "32", nonPIC} };

  struct {
    bool endrungroup;
    const char * iname;
    const char * mrname;
    const char * isoname;
	bool serialize_enable;
    const char * ilabel;
  } test_params[] = { {true, "fucompp", "fucompp", "fucompp.dll", false, "{test: fucompp, mutator: fucompp, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}"},
 {true, "stack_mov_32_bit", "stack_mov_32_bit", "stack_mov_32_bit.dll", false, "{test: stack_mov_32_bit, mutator: stack_mov_32_bit, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}"},
 {true, "mov_size_details", "mov_size_details", "mov_size_details.dll", false, "{test: mov_size_details, mutator: mov_size_details, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: createProcess}"},
 {true, "test_instruction_read_write", "test_instruction_read_write", "test_instruction_read_write.dll", false, "{test: test_instruction_read_write, mutator: test_instruction_read_write, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}"},
 {true, "test_instruction_farcall", "test_instruction_farcall", "test_instruction_farcall.dll", false, "{test: test_instruction_farcall, mutator: test_instruction_farcall, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}"},
 {true, "test_instruction_bind_eval", "test_instruction_bind_eval", "test_instruction_bind_eval.dll", false, "{test: test_instruction_bind_eval, mutator: test_instruction_bind_eval, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}"},
 {true, "power_decode", "power_decode", "power_decode.dll", false, "{test: power_decode, mutator: power_decode, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}"},
 {true, "power_cft", "power_cft", "power_cft.dll", false, "{test: power_cft, mutator: power_cft, grouped: false, pic: none, start_state: stopped, format: dynamicMutatee, process_mode: None, abi: 32, thread_mode: None, mutateeruntime: no_launch, platmode: NONE, mutatee: none, mutatorstart: local, optimization: none, mutateestart: local, compiler: , run_mode: disk}"} };

  int tp_index = 0;
  for (int i = 0; i < 8; i++) {
    test_count = 0;
    rg = new RunGroup(rungroup_params[i].mutatee_name, rungroup_params[i].state_init, rungroup_params[i].attach_init, 
			rungroup_params[i].ex, rungroup_params[i].module, rungroup_params[i].pic, rungroup_params[i].compiler,
			rungroup_params[i].optimization, rungroup_params[i].abi, "NONE");
    
    for (; tp_index < 8; tp_index++) {
      rg->tests.push_back(new TestInfo(test_count++, 
                                       test_params[tp_index].iname, 
                                       test_params[tp_index].mrname, 
                                       test_params[tp_index].isoname, 
                                       test_params[tp_index].serialize_enable, 
                                       test_params[tp_index].ilabel));
      if (test_params[tp_index].endrungroup == true) break;
    } 
    // Increment the test_params index (tp_index) past the endgroup. 
    tp_index++;
    rg->index = group_count++;
    tests.push_back(rg);
  }
}

// Now we insert the test lists into the run groups
void initialize_mutatees_proccontrol(std::vector<RunGroup *> &tests) {
	unsigned int group_count = 0;
	// Keep track of which element each test is, for later use with the resumelog
	unsigned int test_count;
	RunGroup *rg;
struct {

    char * mutatee_name;
    start_state_t state_init;
    create_mode_t attach_init;
    bool ex;
    bool presencevar;
    char* module;
    char* compiler;
    char* optimization;
    char* abi;
    test_pictype_t pic;
  } rungroup_params[] = { {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_launch_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_cont_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_breakpoint_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_library_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_addlibrary_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_singlestep_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_thread_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_groups_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_stat_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_irpc_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_temp_detach_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_terminate_stopped_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, CREATE, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC},
 {"pc_mem_perm_mutatee_solo_VC++_32_none.exe", SELFATTACH, USEATTACH, true, true, "proccontrol", "VC++", "none", "32", nonPIC} };

  struct {
    bool endrungroup;
    const char * iname;
    const char * mrname;
    const char * isoname;
	bool serialize_enable;
    const char * ilabel;
  } test_params[] = { {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_launch", "pc_launch", "pc_launch.dll", false, "{test: pc_launch, mutator: pc_launch, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_launch, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread_cont", "pc_thread_cont", "pc_thread_cont.dll", false, "{test: pc_thread_cont, mutator: pc_thread_cont, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread_cont, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_breakpoint", "pc_breakpoint", "pc_breakpoint.dll", false, "{test: pc_breakpoint, mutator: pc_breakpoint, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_breakpoint, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_library", "pc_library", "pc_library.dll", false, "{test: pc_library, mutator: pc_library, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_library, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_addlibrary", "pc_addlibrary", "pc_addlibrary.dll", false, "{test: pc_addlibrary, mutator: pc_addlibrary, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_addlibrary, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_singlestep", "pc_singlestep", "pc_singlestep.dll", false, "{test: pc_singlestep, mutator: pc_singlestep, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_singlestep, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_thread", "pc_thread", "pc_thread.dll", false, "{test: pc_thread, mutator: pc_thread, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_thread, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_groups", "pc_groups", "pc_groups.dll", false, "{test: pc_groups, mutator: pc_groups, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_groups, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_stat", "pc_stat", "pc_stat.dll", false, "{test: pc_stat, mutator: pc_stat, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_stat, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_irpc", "pc_irpc", "pc_irpc.dll", false, "{test: pc_irpc, mutator: pc_irpc, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_irpc, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_detach", "pc_detach", "pc_detach.dll", false, "{test: pc_detach, mutator: pc_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_temp_detach", "pc_temp_detach", "pc_temp_detach.dll", false, "{test: pc_temp_detach, mutator: pc_temp_detach, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_temp_detach, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate", "pc_terminate", "pc_terminate.dll", false, "{test: pc_terminate, mutator: pc_terminate, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_terminate_stopped", "pc_terminate_stopped", "pc_terminate_stopped.dll", false, "{test: pc_terminate_stopped, mutator: pc_terminate_stopped, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_terminate_stopped, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: no_launch, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: createProcess}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: MultiThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: MultiProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"},
 {true, "pc_mem_perm", "pc_mem_perm", "pc_mem_perm.dll", false, "{test: pc_mem_perm, mutator: pc_mem_perm, grouped: false, pic: none, start_state: selfattach, format: dynamicMutatee, process_mode: SingleProcess, abi: 32, thread_mode: SingleThreaded, mutateeruntime: pre, platmode: NONE, mutatee: pc_mem_perm, mutatorstart: local, optimization: none, mutateestart: local, compiler: VC++, run_mode: useAttach}"} };

  int tp_index = 0;
  for (int i = 0; i < 240; i++) {
    test_count = 0;
    rg = new RunGroup(rungroup_params[i].mutatee_name, rungroup_params[i].state_init, rungroup_params[i].attach_init, 
			rungroup_params[i].ex, rungroup_params[i].module, rungroup_params[i].pic, rungroup_params[i].compiler,
			rungroup_params[i].optimization, rungroup_params[i].abi, "NONE");
    
    for (; tp_index < 240; tp_index++) {
      rg->tests.push_back(new TestInfo(test_count++, 
                                       test_params[tp_index].iname, 
                                       test_params[tp_index].mrname, 
                                       test_params[tp_index].isoname, 
                                       test_params[tp_index].serialize_enable, 
                                       test_params[tp_index].ilabel));
      if (test_params[tp_index].endrungroup == true) break;
    } 
    // Increment the test_params index (tp_index) past the endgroup. 
    tp_index++;
    rg->index = group_count++;
    tests.push_back(rg);
  }
}
